
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/poll.h>
#include <asm/io.h>
#include <asm/uaccess.h>


#define DEB(f)


#define RTC_BASE  0x70
#define RTC_WIDTH 0x02
#define RTC_IRQ   0x08

static const char devname[]="Linux Thingy";

static struct file_operations fops;
static spinlock_t rtc_lock = SPIN_LOCK_UNLOCKED;
static size_t global_poll_flag; 
static wait_queue_head_t pwait;
extern void rtc_isr(void);
extern void init_rtc(void);
extern void stop_rtc(void);
extern void clear_rtc(void);
extern long long rtc_tick;

static void isr(int a, void *b, struct pt_regs *c)
{
    spin_lock(&rtc_lock);
    rtc_isr();
    global_poll_flag = POLLIN;
    wake_up_interruptible(&pwait);
    spin_unlock(&rtc_lock);
}

static size_t poll(struct file *fp, struct poll_table_struct *wait)
{
    size_t poll_flag;
    size_t flags;
    DEB(printk(KERN_INFO"%s : poll() called\n", devname));
    poll_wait(fp, &pwait, wait);
    DEB(printk(KERN_INFO"%s : poll() returned\n", devname));
    spin_lock_irqsave(&rtc_lock, flags);
    poll_flag = global_poll_flag;
    global_poll_flag = 0;
    spin_unlock_irqrestore(&rtc_lock, flags);
    return poll_flag;
}

static int read(struct file *fp, char *buf, size_t count, loff_t *ppos)
{
    long long tmp;
    size_t flags;
    spin_lock_irqsave(&rtc_lock, flags);
    tmp = rtc_tick;
    spin_unlock_irqrestore(&rtc_lock, flags);
    if(copy_to_user(buf, &tmp, sizeof(tmp)))
        return -EFAULT;
    return sizeof(tmp);
}

static int open(struct inode *ip, struct file *fp)
{
    size_t flags;
    MOD_INC_USE_COUNT;
    spin_lock_irqsave(&rtc_lock, flags);
    clear_rtc();
    global_poll_flag = 0;
    spin_unlock_irqrestore(&rtc_lock, flags);
    return 0;
} 
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
static int close(struct inode *ip, struct file *fp)
{
    MOD_DEC_USE_COUNT;
    return 0;
} 
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   Initialize the module and return 0 if everything is okay.
 *   Return a negative error-code if not.
 */
int __init init_module()
{
    int result;
    if(request_region(RTC_BASE, RTC_WIDTH, devname) == NULL)
    {
         printk(KERN_CRIT"%s : Can't allocate I/O ports\n", devname);
         return -ENODEV;
    }
    if((result = request_irq(RTC_IRQ, isr, SA_INTERRUPT, devname, NULL)) < 0)
    {
         release_region(RTC_BASE, RTC_WIDTH);
         printk(KERN_CRIT"%s : Can't allocate IRQ%d\n", devname, RTC_IRQ);
         return result;
    } 
    memset(&fops, 0x00, sizeof(fops));
    init_waitqueue_head(&pwait);
    fops.owner   = THIS_MODULE;
    fops.open    = open;
    fops.release = close;
    fops.poll    = poll;
    fops.read    = read;
    if((result = register_chrdev(MAJOR_NR, devname, &fops)) < 0)
    {
         printk(KERN_CRIT"%s : Can't register character device with "\
         "major number %d\n", devname, MAJOR_NR);
         return result;
    }
    init_rtc();
    printk(KERN_INFO"%s : initialization complete\n", devname);
    return 0;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  Cleanup the module in preparation for removal.
 */
void cleanup_module()
{
    int ret;
    stop_rtc();
    free_irq(RTC_IRQ, NULL);
    if((ret = unregister_chrdev(MAJOR_NR, devname)) < 0)
    {
        printk(KERN_ALERT"%s : Can't unregister major number %d (%d)\n", devname, MAJOR_NR, ret);
        return;
    } 
    release_region(RTC_BASE, RTC_WIDTH);
    printk(KERN_INFO"%s : module removed\n", devname);
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
